# -*- coding: utf-8 -*-
# Copyright 2009 Intevation GmbH, Germany, <info@intevation.de>
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

IDENTIFIER_SEPARATOR = u':'


class InstanceComponent(object):

    def __init__(self, meta=None):
        self.meta = meta
        self.parent = None
        self.tree = None
        self.identifier = None

    def isRepeat(self):
        return False

    def isLeaf(self):
        return False

    def getParent(self):
        return self.parent

    def getMeta(self):
        return self.meta

    def getInstanceTree(self):
        if self.tree:
            return self.tree
        cur = self.parent
        tree = None
        while cur:
            tree = cur.getInstanceTree()
            if tree:
                break
            cur = cur.parent
        self.tree = tree
        return tree

    def getIdentifier(self):
        # TODO: The following line prevents RG from proper working, and is
        # therefor commented out. Can not find any sinde effects yet
        #if self.identifier: return self.identifier

        path = [self.getMeta().getName()]
        cur = self
        while cur:
            if cur.isRepeat():
                path.append(str(cur.getInternalIdentifier()))
            cur = cur.parent
        identifer = IDENTIFIER_SEPARATOR.join(path)
        self.identifer = identifer
        return identifer

    def getInternalIdentifier(self):
        return None

    def hasErrors(self):
        '''Returns True if the fields has an error'''
        tree = self.getInstanceTree()
        if tree:
            return tree.hasErrors(self.getIdentifier())
        return False

    def getErrors(self):
        '''Returns Error message of the field as str'''
        tree = self.getInstanceTree()
        if tree:
            return tree.getErrors(self.getIdentifier())
        return None

    def hasWarning(self):
        '''Returns True if the fields has a warning'''
        tree = self.getInstanceTree()
        if tree:
            return tree.hasWarning(self.getIdentifier())
        return False

    def getWarning(self):
        '''Returns Warning message of the field as str'''
        tree = self.getInstanceTree()
        if tree:
            return tree.getWarning(self.getIdentifier())
        return None


class InstanceLeaf(InstanceComponent):

    def __init__(self, meta=None, value=None):
        InstanceComponent.__init__(self, meta)
        self.value = value

    def getValue(self):
        return self.value

    def isLeaf(self):
        return True


class InstanceNode(InstanceComponent):

    def __init__(self, meta=None, children=None):
        InstanceComponent.__init__(self, meta)
        self.children = children

    def getChildren(self):
        return self.children

    def hasErrors(self):
        my_ids = (IDENTIFIER_SEPARATOR +
                  self.getIdentifier().split(IDENTIFIER_SEPARATOR, 1)[1])
        tree = self.getInstanceTree()
        find = tree.getFormEd().findByName
        my_meta = self.getMeta()
        for identifier in tree.getErrors().iterkeys():
            iname, ids = identifier.split(IDENTIFIER_SEPARATOR, 1)
            if not (IDENTIFIER_SEPARATOR + ids).endswith(my_ids):
                continue
            meta = find(iname)
            while meta:
                if meta == my_meta:
                    return True
                meta = meta.parent
        return False

    def getErrors(self):
        my_ids = (IDENTIFIER_SEPARATOR +
                  self.getIdentifier().split(IDENTIFIER_SEPARATOR, 1)[1])
        errors = []
        tree = self.getInstanceTree()
        find = tree.getFormEd().findByName
        my_meta = self.getMeta()
        for identifier, error_item in tree.getErrors().iteritems():
            iname, ids = identifier.split(IDENTIFIER_SEPARATOR, 1)
            if not (IDENTIFIER_SEPARATOR + ids).endswith(my_ids):
                continue
            meta = find(iname)
            while meta:
                if meta == my_meta:
                    errors.append(error_item)
                    break
                meta = meta.parent
        return errors


class PageInstanceNode(InstanceNode):

    pass


class RepeatInstanceContainerNode(InstanceNode):

    def createInstance(self, uuid=None):
        return None

    def deleteInstance(self, instance_node):
        pass


class RepeatInstanceNode(InstanceNode):

    def __init__(self, meta=None, children=None, identifier=None):
        InstanceNode.__init__(self, meta, children)
        self.identifier = identifier

    def isRepeat(self):
        return True


class RootNode(InstanceNode):

    def __init__(self, meta=None, children=None, identifier=None):
        InstanceNode.__init__(self, meta, children)
        self.identifier = identifier

    def getIdentifier(self):
        return self.identifier

    def isRepeat(self):
        return True


class SetDataException(Exception):

    def __init__(self, errors):
        Exception.__init__(self)
        self.errors = errors


class InstanceTree(object):

    def __init__(self, formed, root=None):
        self.formed = formed
        if root is None:
            root = RootNode()
        self.root = root

    def getFormEd(self):
        return self.formed

    def getItem(self, identifier, name=None):
        return None

    def getAllItemsForName(self, name):
        return None

    def hasErrors(self, identifier=None):
        return None

    def getErrors(self, identifier=None):
        return None

    def commit(self):
        return None

    def dryConvert(self, data):
        '''data is dict'''
        pass

    def setData(self, data):
        pass

    def setDataNoChecks(self, data):
        pass

    def getRootNode(self):
        return self.root


# vim:set ts=4 sw=4 si et sta sts=4 enc=utf-8 fenc=utf-8 :
