# -*- coding: latin-1 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
import os
from datetime import date

from formed.meta.nodecomponents import Node, Leaf, CaseNode, \
     RootNode, VISIT_IGNORE_CHILDREN, VISIT_CONTINUE
from formed.meta.exprtree import Expr


def int_or_none(string_value):
    """Convert the string parameter to int. If the string is empty, return None.
    """
    if not string_value:
        value = None
    else:
        value = int(string_value)
    return value

def float_or_none(string_value):
    """Convert the string parameter to flat. If the string is empty, return None.
    """
    if not string_value:
        value = None
    else:
        value = float(string_value)
    return value


class RelevantVisitor(object):

    def __init__(self, callback=None):
        self.callback = callback

    def visitor(self, nc):
        if isinstance(nc, (RadioNode, ChoiceNode)):
            self.callback(nc)
            return VISIT_IGNORE_CHILDREN
        if isinstance(nc, LogbookNode):
            # eat it!
            return VISIT_IGNORE_CHILDREN
        if isinstance(nc, (CaseNode, RootNode)):
            return VISIT_CONTINUE
        self.callback(nc)
        return VISIT_CONTINUE


class WidgetCollector(object):

    def __init__(self, root=None):
        self.widgets = []
        self.root = root

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, (RadioNode, ChoiceNode)):
                self.widgets.append(nc)
                return VISIT_IGNORE_CHILDREN
            if isinstance(nc, RepeatNode):
                return self.handleRepeatNode(nc)
        elif (isinstance(nc, Leaf)
              and not isinstance(nc, (RuleLeaf, ExternalChoiceListLeaf,
                                      InfoLeaf))):
            self.widgets.append(nc)

        return VISIT_CONTINUE

    def handleRepeatNode(self, repeat):
        if repeat != self.root:
            return VISIT_IGNORE_CHILDREN
        return VISIT_CONTINUE


class ConditionalWidgetCollector(object):

    def __init__(self, evaluator=lambda expr: True):
        self.widgets = []
        self.evaluator = evaluator

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, ConditionalNode):
                expr = nc.getExpr()
                if expr and not self.evaluator(expr):
                    return VISIT_IGNORE_CHILDREN
            elif isinstance(nc, (RadioNode, ChoiceNode)):
                self.widgets.append(nc)
                return VISIT_IGNORE_CHILDREN
            elif isinstance(nc, GroupNode) and nc.isRepeat() and nc != self:
                return VISIT_IGNORE_CHILDREN
        elif (isinstance(nc, Leaf)
              and not isinstance(nc, (RuleLeaf, ExternalChoiceListLeaf,
                                      InfoLeaf))):
            self.widgets.append(nc)

        return VISIT_CONTINUE


class FilteredNodeComponentProxy:

    def __init__(self, params):
        self.params = params

    def set(self, nc):
        self.nc = nc

    def getName(self):
        if not self.nc:
            return ""
        return self.nc.getName()

    def allWidgets(self):
        if self.nc:
            print >> sys.stderr, self.nc
            return [w for w in self.nc.allWidgets()
                    if w.getName() in self.params]
        return []


class PageNode(Node):

    def __init__(self):
        Node.__init__(self)
        self.attributes["invisible"] = ""

    def __repr__(self):
        return "<PageNode 0x%x name=%r>" % (id(self), self.getName())

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def allWidgets(self):
        collector = WidgetCollector()
        self.visit(collector.visitor)
        return collector.widgets


class LogbookNode(Node):

    pass


class GroupNode(Node):

    def __init__(self):
        Node.__init__(self)
        self.attributes["repeat"] = ""
        self.attributes["invisible"] = ""

    def __repr__(self):
        return "<GroupNode 0x%x name=%r>" % (id(self), self.getName())

    def isRepeat(self):
        return self.getAttribute("repeat") and True or False

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def allWidgets(self):
        collector = WidgetCollector()
        self.visit(collector.visitor)
        return collector.widgets


class RepeatNode(Node):

    def __init__(self):
        Node.__init__(self)
        self.attributes["min"] = ""
        self.attributes["max"] = ""
        self.attributes["digest"] = ""

    def __repr__(self):
        return ("<RepeatNode 0x%x name=%r min=%r max=%r>"
                % (id(self), self.getName(), self.getMin(), self.getMax()))

    def getMin(self):
        return self.getAttribute("min")

    def getMax(self):
        return self.getAttribute("max")

    def getDigest(self):
        return self.getAttribute("digest")


class ConditionalNode(Node):

    def __init__(self):
        Node.__init__(self)
        self.attributes["expr"] = ""
        self.attributes["invisible"] = ""
        self.expr = None

    def __repr__(self):
        return ("<ConditionalNode 0x%x name=%r expr=%r invisible=%r>"
                % (id(self), self.getName(), self.expr.expr,
                   self.isInvisible()))

    def setAttribute(self, attrib, value):
        if attrib == "expr":
            newExpr = Expr(value)
            newExpr.compile()
            self.expr = newExpr
        Node.setAttribute(self, attrib, value)

    def setExpr(self, expr):
        self.setAttribute('expr', expr)

    def getExpr(self):
        return self.expr

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False


class MatrixNode(GroupNode):

    def __repr__(self):
        return "<MatrixNode 0x%x name=%r>" % (id(self), self.getName())


class RadioNode(Node):

    def __init__(self):
        Node.__init__(self)
        self.attributes["selected"] = ""

    def __repr__(self):
        return ("<RadioNode 0x%x name=%r selected=%r>"
                % (id(self), self.getName(), self.getSelected()))

    def firstFindDefault(self):
        value = None
        if self.children:
            for child in self.children:
                if isinstance(child, ExternalChoiceListLeaf):
                    found, v = child.find_default()
                    if found:
                        value = v
                        break
                elif isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v:
                            value = int(v)
                        break

        self.found_default = value
        return Node.firstFindDefault(self)

    def getSelected(self):
        return self.getAttribute("selected")


class BoolLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.attributes["checked"] = ""
        self.attributes["invisible"] = ""

    def __repr__(self):
        return ("<BoolLeaf 0x%x name=%r value=%r checked=%r>"
                % (id(self), self.getName(), self.getValue(),
                   self.getChecked()))

    def getChecked(self):
        return self.getAttribute("checked")

    def getValue(self):
        """Return the value associated with the bool leaf as an int.
        If no value has been set, return None.
        """
        return int_or_none(Leaf.getValue(self))

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False


class ExternalChoiceListLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.choice = None
        self.path = None

    def __repr__(self):
        return ("<ExternalChoiceListLeaf 0x%x name=%r value=%r>"
                % (id(self), self.getName(), self.getValue()))

    def setLoadContext(self, path):
        self.path = os.path.dirname(path)

    def find_default(self):
        found = False, None
        children = self.getChildren()
        if children:
            for child in children:
                if isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v:
                            found = (True, int(v))
                        break
        return found

    def getChildren(self):
        if not self.choice:
            fname = self.getValue()
            if fname:
                if not os.path.isabs(fname) and self.path:
                    fname = os.path.join(self.path, fname)
                # TODO: This loading code should be else where!
                from formed.meta.io.document import openDocument
                document = openDocument(fname)
                self.choice = document.findByClassAndName(ChoiceNode)
                if not self.choice:
                    raise RuntimeError("No ChoiceNode in %s" % fname)
        return self.choice.children


class InfoLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.attributes["invisible"] = ""

    def __repr__(self):
        return "<InfoLeaf 0x%x name=%r>" % (id(self), self.getName())

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False


class TextAreaLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.attributes["cols"] = ""
        self.attributes["rows"] = ""

    def __repr__(self):
        return ("<TextAreaLeaf 0x%x name=%r cols=%r rows=%r>"
                % (id(self), self.getName(), self.getCols(), self.getRows()))

    def firstFindDefault(self):
        value = self.getValue()
        if not value:
            value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

    def getCols(self):
        return self.getAttribute("cols")

    def getRows(self):
        return self.getAttribute("rows")


class TextLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.attributes["size"] = ""
        self.attributes["maxlength"] = ""

    def __repr__(self):
        return ("<TextLeaf 0x%x name=%r size=%r maxlength=%r>"
                % (id(self), self.getName(), self.getSize(),
                   self.getMaxLength()))

    def firstFindDefault(self):
        value = self.getValue()
        if not value:
            value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

    def getSize(self):
        return self.getAttribute("size")

    def getMaxLength(self):
        """Return the maximum length of the field in characters.
        If the length is unlimited, return None.
        """
        return int_or_none(self.getAttribute("maxlength"))


class ChoiceNode(Node):

    def __init__(self):
        Node.__init__(self)
        self.attributes["size"] = "1"
        self.attributes["multiple"] = ""

    def __repr__(self):
        return "<ChoiceNode 0x%x name=%r>" % (id(self), self.getName())

    def firstFindDefault(self):
        value = None
        if self.children:
            for child in self.children:
                if isinstance(child, ExternalChoiceListLeaf):
                    found, v = child.find_default()
                    if found:
                        value = v
                        break
                elif isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v:
                            value = int(v)
                        break

        self.found_default = value
        return Node.firstFindDefault(self)

    def getSize(self):
        return self.getAttribute("size")

    def getMultiple(self):
        return self.getAttribute("multiple")

    def getSelected(self):
        return self.getAttribute("selected")

    def getValueMapping(self):
        """Returns a dictionary mapping the values of each option in the choice
        list to its description"""
        mapping = {}
        for c in self.walk():
            if isinstance(c, BoolLeaf):
                mapping[c.getValue()] = c.getDescription()
        return mapping


class IntLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.attributes["minvalue"] = ""
        self.attributes["maxvalue"] = ""

    def __repr__(self):
        return ("<IntLeaf 0x%x name=%r minvalue=%r maxvalue=%r>"
                % (id(self), self.getName(), self.getMinValue(),
                   self.getMaxValue()))

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = int(value)
            except ValueError:
                print >> sys.stderr, "Broken default for '%s'" % self.getName()
                value = None
        else:
            value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

    def getMinValue(self):
        return int_or_none(self.getAttribute("minvalue"))

    def getMaxValue(self):
        return int_or_none(self.getAttribute("maxvalue"))

class NumericLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.attributes["minvalue"] = ""
        self.attributes["maxvalue"] = ""

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = float(value)
            except ValueError:
                print >> sys.stderr, "Broken default for '%s'" % self.getName()
                value = None
        else:
            value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)

    def getMinValue(self):
        return float_or_none(self.getAttribute("minvalue"))

    def getMaxValue(self):
        return float_or_none(self.getAttribute("maxvalue"))

class DateLeaf(Leaf):

    def __repr__(self):
        return "<DateLeaf 0x%x name=%r>" % (id(self), self.getName())

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = date(*[int(x) for x in value.split("-")])
            except (TypeError, ValueError):
                print >> sys.stderr, "Broken default for '%s'" % self.getName()
                value = None
        else:
            value = None
        self.found_default = value
        return Leaf.firstFindDefault(self)


class RuleLeaf(Leaf):

    def __init__(self):
        Leaf.__init__(self)
        self.attributes["expr"] = ""
        self.attributes["mark"] = ""
        self.expr = None

    def __repr__(self):
        return ("<RuleLeaf 0x%x name=%r expr=%r mark=%r>"
                % (id(self), self.getName(), self.expr.expr,
                   self.getMark()))

    def shallowCopy(self):
        r = Leaf.shallowCopy(self)
        if self.expr:
            r.expr = self.expr.clone()
        else:
            r.expr = None
        return r

    def setAttribute(self, attrib, value):
        if attrib == "expr":
            newExpr = Expr(value)
            newExpr.compile()
            self.expr = newExpr
        Leaf.setAttribute(self, attrib, value)

    def getExpr(self):
        return self.expr

    def getMark(self):
        return self.getAttribute("mark")

    def setMark(self, mark):
        self.setAttribute("mark", mark)

    def isAlwaysAvailable(self):
        """Return whether the rule is always available.
        If a rule is always available, it should be evaluated even if
        the fields it depends on belong to conditionals that are false.
        A rule is always available if and only if it is an error rule as
        determined by the isError method.
        """
        return self.isError()

    def isError(self):
        mark = self.getMark()
        return mark and mark.startswith("error")

    def isWarning(self):
        mark = self.getMark()
        return mark and (mark.startswith("warning")
                         or mark.startswith("notice"))

    def getMarkedName(self):
        mark = self.getMark()
        if mark:
            split_mark = mark.split(":")
            if len(split_mark) == 2:
                return split_mark[-1]
            else:
                raise ValueError("There must be exactly one colon in a"
                                 " mark value")
        return None


# vim:set ts=4 sw=4 si et sta sts=4:
