"""Pylons middleware initialization"""

import logging
import locale

from webob.exc import HTTPMovedPermanently
from beaker.middleware import CacheMiddleware, SessionMiddleware
from paste.cascade import Cascade
from paste.registry import RegistryManager
from paste.urlparser import StaticURLParser
from paste.deploy.converters import asbool
from pylons import config
from pylons.util import  class_name_from_module_name
from pylons.middleware import ErrorHandler, StatusCodeRedirect
from pylons.wsgiapp import PylonsApp
from routes.middleware import RoutesMiddleware

from mpulsweb.config.environment import load_environment
from mpulsweb.config.importer import import_overridable_module


log = logging.getLogger(__name__)


class SimpleExceptionLogger(object):

    """Simply log any exception raised while a request is handled.

    If an exception is raised during a request, the exception is logged
    with log.exception and reraised without any further processing.
    This can be used to make sure that the actual exception information
    is logged even if the more sophisticated error handlers of
    e.g. Pylons fail.  For that use-case it has to be as near to the
    actual application in the application stack as possible.
    """

    def __init__(self, application):
        self.application = application

    def __call__(self, *args, **kw):
        try:
            return self.application(*args, **kw)
        except:
            log.exception("Exception while handling request")
            raise


class MyPylonsApp(PylonsApp):

    def find_controller(self, controller):
        """Override this method to look for the controller in two places.

        First, the controller is searched in the MPuls specialization's
        controllers package, and if it cannot be found there, it's
        imported from the mpulsweb.
        """
        # Check to see if we've cached the class instance for this name
        if controller in self.controller_classes:
            return self.controller_classes[controller]

        # Hide the traceback here if the import fails (bad syntax and such)
        __traceback_hide__ = 'before_and_this'

        module = import_overridable_module("controllers."
                                           + controller.replace('/', '.'))

        if hasattr(module, '__controller__'):
            mycontroller = getattr(module, module.__controller__)
        else:
            module_name = controller.split('/')[-1]
            class_name = class_name_from_module_name(module_name) + 'Controller'
            if self.log_debug:
                log.debug("Found controller, module: %r, class: %r",
                          module, class_name)
            mycontroller = getattr(module, class_name)
        self.controller_classes[controller] = mycontroller
        return mycontroller


def make_app(global_conf, full_stack=True, static_files=True, **app_conf):
    """Create a Pylons WSGI application and return it

    ``global_conf``
        The inherited configuration for this application. Normally from
        the [DEFAULT] section of the Paste ini file.

    ``full_stack``
        Whether this application provides a full WSGI stack (by default,
        meaning it handles its own exceptions and errors). Disable
        full_stack when this application is "managed" by another WSGI
        middleware.

    ``static_files``
        Whether this application serves its own static files; disable
        when another web server is responsible for serving them.

    ``app_conf``
        The application's local configuration. Normally specified in
        the [app:<name>] section of the Paste ini file (where <name>
        defaults to main).

    """
    # Configure the Pylons environment
    load_environment(global_conf, app_conf)

    # Set locale
    locale.setlocale(locale.LC_ALL, global_conf.get('locale', 'de_DE'))
    # The Pylons WSGI app
    app = MyPylonsApp()

    # Pylon's exception handler (at least the debug handler) sometimes
    # has problems with exceptions whose messages are byte-strings
    # containing non-ascii characters, which can happen when
    # e.g. psycopg2 is used in a German locale.  To avoid loss of
    # information, we put the SimpleExceptionLogger immediately around
    # the pylons app.
    if config.get("mpuls.app.simple-exception-handler", False):
        log.debug("Using MPuls SimpleExceptionLogger")
        app = SimpleExceptionLogger(app)
    else:
        log.debug("Not using MPuls SimpleExceptionLogger")

    # Routing/Session/Cache Middleware
    app = RoutesMiddleware(app, config['routes.map'])
    app = SessionMiddleware(app, config)
    app = CacheMiddleware(app, config)

    # CUSTOM MIDDLEWARE HERE (filtered by error handling middlewares)

    if asbool(full_stack):
        # Handle Python exceptions
        app = ErrorHandler(app, global_conf, **config['pylons.errorware'])

        # Display error documents for 401, 403, 404 status codes (and
        # 500 when debug is disabled)
        if asbool(config['debug']):
            app = StatusCodeRedirect(app)
        else:
            app = StatusCodeRedirect(app, [400, 401, 403, 404, 500])

    # Establish the Registry for this application
    app = RegistryManager(app)

    if asbool(static_files):
        # Serve static files
        static_apps = [StaticURLParser(path)
                       for path in config['pylons.paths']['static_files']]
        app = Cascade(static_apps + [app])

    return app


def make_redirector(global_conf, **app_conf):
    """Create a WSGI Application that permanently redirects to a specific URL.
    The URL can be specified with the location setting in the configuration.

    This is useful for e.g. a composite application with two WSGI
    applications under different prefixes.  In this case one may want to
    specify a default application to use when accessing the root URL of
    the server.
    """
    return HTTPMovedPermanently(location=app_conf["location"])
