from StringIO import StringIO
import logging
import itertools
import zipfile
import stat
from collections import defaultdict

import pyExcelerator

from formed.meta.data import PageNode, RepeatNode
from formed.meta.structure import StructureWidgetCollector

from mpulsweb.lib.base import g
from mpulsweb.lib.translation import _

log = logging.getLogger(__name__)


# Helpers to work around a problem in pyExcelerator.  pyExcelerator can
# only write to a file that it opens itself using the filename passed to
# the Workbook.save method.  We want the output in a StringIO object,
# though.  To achieve that we override the built-in "file" with the
# fake_file function defined below in pyExcelerator.CompoundDoc.

class MyStringIO(StringIO):

    def close(self):
        pass

    def real_close(self):
        StringIO.close(self)


class FakeStr(str):

    def __init__(self, buf):
        self.buf = buf

def fake_file(x, *args):
    """Wrapper around the built-in 'file' which treats FakeStr specially.
    If the filename argument is an instance of FakeStr, simply return
    its buf attribute.  Otherwise delegate to file.  This allows us to
    pass e.g. a MyStringIO instance to pyExcelerator as a 'filename'
    when we wrap it in a FakeStr.
    """
    return isinstance(x, FakeStr) and x.buf or file(x, *args)

pyExcelerator.CompoundDoc.file = fake_file

#
# Export base classes
#

class Export:

    pass


class Table(object):

    """A single Table for tabular output formats like XLS or CSV"""

    def __init__(self, title, columns):
        """Initialize the table.
        The parameter title is a string with the title of the table and
        will usually be taken from the page or repeat group that is
        represented by the table.  The parameter columns is a list
        defining the column names"""
        self.title = title
        self.columns = columns
        self.rows = []

    def append(self, row):
        """Append a row to the table.
        The parameter row should be a dictionary mapping column names to
        values.
        """
        self.rows.append(row)


class Container(object):

    """A container of multiple Table instances"""

    def __init__(self, meta_tree, selection, table_factory,
                 one_table_per_page=False):
        """Initialize the container with a FormEd meta tree and the selection.
        """
        self.meta_tree = meta_tree
        self.table_factory = table_factory
        self.tables = dict()
        self.field_name_to_table_name = dict()
        self.table_order = []
        self.one_table_per_page = one_table_per_page
        self.create_tables(selection)

    def create_tables(self, selection):
        table_columns = dict()
        repeat_groups = set()
        titles = dict()
        table_order = []
        if not self.one_table_per_page:
            table_order.append("master")
        for page in self.meta_tree.findAllByClass(PageNode):
            # always export the invisible pages as they cannot be
            # explicily selected by the user and they usually contain
            # important information like the current phase.
            if page.getName() not in selection and not page.isInvisible():
                continue

            collector = StructureWidgetCollector()
            page.visit(collector.visitor)
            for nc in collector.widgets:
                if self.one_table_per_page:
                    table_name = page.getName()
                    table_title = page.getDescription()
                else:
                    table_name = "master"
                    table_title = "Master"
                for p in nc.pathToRoot():
                    if isinstance(p, RepeatNode):
                        table_name = p.getName()
                        table_title = p.getDescription()
                        repeat_groups.add(table_name)
                        break

                if table_name not in table_columns:
                    table_order.append(table_name)
                table_columns.setdefault(table_name, []).append(nc.getName())
                titles[table_name] = table_title

        for name, columns in table_columns.iteritems():
            extra_columns = ["id"]
            if name in repeat_groups:
                extra_columns.append("master_id")
            self.tables[name] = self.table_factory(titles[name],
                                                   extra_columns + columns)
            for column_name in columns:
                self.field_name_to_table_name[column_name] = name

        self.table_order = table_order

    def fill(self, tree):
        """Fill the tables with the contents of the ElementTree tree."""
        self.repeat_group_ids = itertools.count()
        for case_id, case in enumerate(tree.getiterator('master')):
            self.fill_from_node(case, case_id)

    def fill_from_node(self, node, node_id, parent_id=None):
        rows = defaultdict(lambda: dict(id=node_id))

        if parent_id is not None:
            rows[node.tag]["master_id"] = parent_id

        for child in node.getchildren():
            name = child.tag
            nc = self.meta_tree.findByName(name)
            if isinstance(nc, RepeatNode):
                self.fill_from_node(child, self.repeat_group_ids.next(),
                                    node_id)
            else:
                table_name = self.field_name_to_table_name.get(name)
                if table_name is not None:
                    rows[table_name][name] = child.text

        for table_name, row in rows.iteritems():
            self.tables[table_name].append(row)


#
# XLS Export
#

class XLSTable(Table):

    def add_to_workbook(self, workbook):
        sheet = workbook.add_sheet(self.title)
        for column, name in enumerate(self.columns):
            sheet.write(0, column, name)

        for row_num, row in enumerate(self.rows):
            for column, name in enumerate(self.columns):
                sheet.write(row_num + 1, column, self.encode(row.get(name)))

    def encode(self, value):
        if value is None:
            return ""
        if isinstance(value, basestring) and len(value) > 32765:
            value = value[:32765]
        return value


class XLSContainer(Container):

    sheet_name_forbidden_characters = dict((ord(c), None) for c in ":\/?*[]")

    max_sheet_name_len = 31


    def write_xls(self, output):
        used_names = set()
        workbook = pyExcelerator.Workbook()
        for name in self.table_order:
            if name in self.tables:
                table = self.tables[name]
                table.title = self.sheet_name_from_title(table.title,
                                                         used_names)
                used_names.add(table.title)
                table.add_to_workbook(workbook)
        workbook.save(FakeStr(output))

    def sheet_name_from_title(self, title, used):
        """Return the name of a worksheet based on the title.
        Worksheet names have to follow the following rules (Excel 2003):

          - They may be at most 31 characters long
          - They must not contain any of the these characters: : \ / ? * [ ]
          - They must not be empty
          - They must be unique in the workbook

        The worksheet name is simply the title, unless one the rules is
        violated.  If a rule is violated, the method logs the
        appropriate warning and tries to correct the problems.  Any
        invalid characters are removed.  Names that are too long are
        shortened.  Empty or duplicate names are replaced with a generic
        but unique name.
        """
        name = title.translate(self.sheet_name_forbidden_characters)
        warnings = []
        if name != title:
            log.warning("Stripped invalid characters from sheet name %r", title)
        if len(name) > self.max_sheet_name_len:
            log.warning("excel sheet name %r is too long; stripping extra"
                        " characters", name)
            # name may end with whitespace after truncation, so strip that, too.
            name = name[:self.max_sheet_name_len].strip()
        if not name or name in used:
            old_name = name
            if not name:
                log.warning("empty name for an excel sheet; selecting unique"
                            " alternative")
                name = _("Table")
            else:
                log.warning("sheet name %r already used; selecting unique"
                            " alternative", name)
            # start counting at 1
            num = len(used) + 1
            while 1:
                numstr = " %d" % num
                candidate = (name[:self.max_sheet_name_len - len(numstr)]
                             + numstr)
                if candidate not in used:
                    name = candidate
                    break
                num += 1
        if name != title:
            log.warning("Using %r instead of %r as excel sheet name", name,
                        title)

        return name



class XLSExport(Export):

    def __init__(self, tree, selection):
        self.tree = tree
        self.xls = XLSContainer(g.formedTree, selection, XLSTable,
                                one_table_per_page=True)

    def export(self):
        self.xls.fill(self.tree)
        try:
            out = MyStringIO()
            self.xls.write_xls(out)
            return out.getvalue()
        finally:
            out.real_close()


#
# CSV Export
#

class CSVTable(Table):

    """A single Table that can be serialized as CSV"""

    def serialize(self, sep):
        """Return the table as a string in CSV format.

        The parameter sep is the separator character.  Rows are
        separated by newlines.  If the separator character occurs in
        values, all occurrences are replaced by spaces or, if the
        separator is the space character, by tabs.
        """
        lines = ["#" + sep.join(self.columns)]
        for row in self.rows:
            lines.append(sep.join(self.encode(row.get(col), sep)
                                  for col in self.columns))
        lines.append("")
        return "\n".join(lines)

    def encode(self, value, sep):
        # first, convert non-strings to strings
        if isinstance(value, int):
            # Simply converting ints to unicode assumes that the
            # separator char is neither a digit nor the minus sign.
            value = unicode(value)
        elif value is None:
            value = ""

        # convert strings if they contain the separator or CR or NL.
        if isinstance(value, basestring):
            rep = "\t" if sep == " " else " "
            value = value.replace(sep, rep)
            value = value.replace("\r\n", rep)
            value = value.replace("\r", rep)
            value = value.replace("\n", rep)
            return value
        else:
            raise TypeError("Cannot handle type %r in CSV export" % type(value))



class CSVContainer(Container):

    """A container of multiple CSVTable instances"""

    def write_zip_file(self, output, encoding="utf-8"):
        """Write the contents of the container to output as a zip-File"""
        def add(name, contents):
            name = name.encode(encoding)
            z.writestr(name, contents.encode(encoding))

            # By default all permission bits are unset when writestr is
            # used so that when unzipping the resulting zip-file, all
            # files are unreadable.  Set default permissions so that the
            # owner may read and write and all other can read.
            info = z.getinfo(name)
            # expression to set unix permissions on external_attr taken
            # from the Python 2.5 zipfile module
            permissions = (stat.S_IRUSR | stat.S_IWUSR
                           | stat.S_IRGRP | stat.S_IROTH)
            info.external_attr = (permissions & 0xFFFF) << 16L

        z = zipfile.ZipFile(output, "w", zipfile.ZIP_DEFLATED)
        sep = "\t"
        add("master.csv", self.tables["master"].serialize(sep))
        for name, table in self.tables.iteritems():
            if name != "master":
                add(name + '.csv', table.serialize(sep))



class CSVExport(Export):

    def __init__(self, tree, selection):
        self.tree = tree
        self.csv = CSVContainer(g.formedTree, selection, CSVTable)

    def export(self):
        self.csv.fill(self.tree)
        s = StringIO()
        self.csv.write_zip_file(s)
        return s.getvalue()
