# Copyright 2007, 2008, 2010 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.

"""Helper functions

Consists of functions to typically be used within templates, but also
available to Controllers. This module is available to both as 'h'.
"""

import sys
import locale
import logging
import re
import datetime
from string import printable
from xml.sax.saxutils import quoteattr

from pylons import c, cache, config, g, request, response, url
from mpulsweb.lib.translation import _

# Use url_for from prior versions
from routes import url_for

from webhelpers.html import escape, HTML, literal, url_escape
from webhelpers.html.tags import *

from formed.meta.data import PageNode 

from mpulsweb.lib.security import hasRole, getKAName

from mpulsweb.lib.helper.filters import shorten, nl_to_br, NA

# normally we would import session from mpulsweb.lib.base, but to avoid
# circular imports we import it from app_globals.
from mpulsweb.lib.app_globals import session


PRINTABLE = frozenset(printable)

log = logging.getLogger(__name__)


# unicode helper
def ensure_unicode(s):
    """Convert s to a unicode object.
    If it's alredy a unicode object, s is returned as is.  Otherwise
    it's converted to a byte-string which is assumed to be in utf-8 and
    then converted to unicode."""
    if isinstance(s, unicode):
        return s
    return unicode(str(s), 'utf-8')

# html helper
def tag(tagname, _close=False,
        _booleans=('checked', 'disabled', 'multiple', 'selected'), **attrs):
    """Create an opening HTML tag with properly escaped attributes.

    The tagname parameter gives the name of the tag.  The attributes and
    their values are given as keyword parameters.  Most attributes have
    string values.  Boolean attributes are treated as described below.
    The values of string attributes are converted to strings if they are
    not aleady strings (byte-strings or unicode) using these rules: None
    is converted to the empty string, other values are converted using
    the str function.  After conversion the values of string attributes
    will be properly escaped as attribute values when inserted into the
    tag.

    Boolean flags are treated differently: if the value of such an
    attribute is true, the attribute is included in the tag without a
    value, if the value is false, the attribute is omitted.  The boolean
    attributes handled in this way are given by the _booleans parameter,
    which is a sequence of attribute names.  Its default value is
    ('checked', 'disabled', 'multiple', 'selected')

    To allow for attribute names that are also python keywords, a single
    trailing underscore character is removed from an attribute name if
    it is preset, so that e.g. the class attribute can be specified
    using the class_ keyword parameters.

    A few attribute names are not valid Python identifiers because they
    contain a minus-sign (e.g. accept-charset).  However, since there
    are no attribute names that contain underscores, underscores in a
    keyword argument will be converted to minus-signs to determine the
    attribute name.  For instance the accept-charset attribute can be
    given as accept_charset to this function.

    Examples:
    >>> from mpulsweb.lib.renderer import tag
    >>> tag("input", type="text", name="title", disabled=False)
    '<input type="text" name="title">'
    >>> 
    """
    out = ["<%s" % tagname]
    for name, value in attrs.items():
        if name.endswith("_"):
            name = name[:-1]
        name = name.replace("_", "-")
        if name in _booleans:
            if value:
                out.append(' %s="%s"' % (name, name))
        else:
            if value is None:
                value = ""
            elif not isinstance(value, basestring):
                value = str(value)
            out.append(' %s="%s"' % (name, escape(value, True)))
    if _close:
        out.append(">")
    else:
        out.append(">")
    return "".join(out)

def render_logbook_typeoptions(item):
    out = []
    for i in item.get('options'):
        if isinstance(i, dict):
            out.append('<optgroup label="%s">' % i.get('label'))
            out.extend(render_logbook_typeoptions(i))
            out.append('</optgroup>')
        else:
            out.append('<option value="%s">%s</option>' % (i[0], i[1]))
    print out
    return out

def convertForDigest(value):
    """Will convert given values to it's string representation for correct
    display in digest. Will localize dates and shown and return string
    'unknown' if it is the unknown value"""
    if isinstance(value, datetime.date):
        if value == datetime.date(1,1,1):
            return _('unknown')  
        else:
            return format_date(value)
    if isinstance(value, int):
        if value == -9999999:
            return _('unknown')  
    if value == None:
        return ""
    return value

# Helper functions used in the UI to show some information like status messages
# or icons

def get_sorted_db_from_selectionlist():
    """Returns a sorted list of available dbs"""
    d = get_db_selectionlist()
    s = {} 
    name = re.compile('\D*')
    for k in d.keys():
        i = int(name.split(k)[1])
        s[i] = k

    l = []
    for id in sorted(s.keys()):
        l.append(s.get(id))
    return l

def get_db_selectionlist():
    """Returns a dictionarie used for listing available db on the loginscreen."""
    dblist = g.mpuls_dblist.get_list()
    return dblist

def status_message(s):
    return "%s: %s" % (format_datetime(datetime.datetime.now()), s)

def get_logouttime():
    """Estimate log-out time and return it as a formatted string (HH:MM).
    The estimated logout time is two minutes before the time the session
    actually expires to make sure it's earlier.
    """
    timeout = int(config.get("beaker.session.timeout"))
    return format_time(datetime.datetime.now()
                       + datetime.timedelta(seconds=timeout - 120), True)

def getHelp(help_id):
    '''
    Returns a HTML link with a help icon if help is present for help_id. Else
    return empty string
    '''
    out = []
    help = g.helpData.getHelp(help_id)
    if help:
        out.append(tag("a", href=url_for(controller='/annotations',
                                         action='help', id=help_id),
                       target="_blank"))
        out.append(tag("img", src=url_for('/images/icons/formular/help.png'),
                       alt="help"))
        out.append('</a>')
    return literal("".join(out))

def get_app_title():
    return config.get('mpuls.app.title')


# Helper functions to get some information about the logged in user

def getRole():
    try:
        user = session['USER_AUTHORIZED']
        return user.getRole()
    except KeyError:
        return ''

def getUserLastname():
    try:
        return session['USER_AUTHORIZED'].last_name
    except KeyError:
        return ''

def getUserFullname():
    try:
        user = session['USER_AUTHORIZED']
        return "%s, %s" % (user.last_name, user.first_name)
    except KeyError:
        return ''

def getLogin():
    try:
        user = session['USER_AUTHORIZED']
        return user.login
    except KeyError:
        return ''

def getUserId():
    try:
        user = session['USER_AUTHORIZED']
        return user.id
    except KeyError:
        return ''

# Format functions

def format_number(number, precision=2):
    fmt = "%%.%sf" % precision
    retval = locale.format(fmt, (number), 1)
    return retval

def format_date(dt):
    if dt is None:
        return ''
    try:
        cs = locale.nl_langinfo(locale.CODESET)
        return unicode(dt.strftime(locale.nl_langinfo(locale.D_FMT)), cs)
    except Exception, err:
        log.error(err)
        return dd_mm_YYYY(dt)

def format_time(dt, approximate=False):
    """
    Format time with the correct locale. If the optional parameter "approximate"
    is set, the minutes will be rounded to the previous multiple of 5
    """
    if dt is None:
        return ''
    try:
        cs = locale.nl_langinfo(locale.CODESET)
        if approximate:
            dt = dt.replace(minute = int(5 * (dt.minute/5)))
        return unicode(dt.strftime('%H:%M'), cs)
    except Exception, err:
        log.error(err)
        return ''

def format_datetime(dt):
    if dt is None:
        return ''
    try:
        t = dt.strftime(locale.nl_langinfo(locale.D_T_FMT))
        cs = locale.nl_langinfo(locale.CODESET)
        return unicode(dt.strftime(locale.nl_langinfo(locale.D_T_FMT)), cs)
    except Exception, err:
        log.error(err)
        return dd_mm_yyyy_HH_MM(dt)

def timedelta_in_minutes(ti):
    if ti is None:
        return None
    return round(float(ti.days * (24*60)) + ti.seconds / 60.0)

def timedelta_in_weeks(ti, empty=None):
    if ti is None:
        return empty
    return format_number(float(ti.days) / 7.0
                         + ti.seconds / (60.0 * 24.0 * 7.0))

def timedelta_in_years(ti, empty=None, precision=None):
    if ti is None:
        return empty
    return format_number(float(ti.days) / 365.0 
                         + ti.seconds / (60.0 * 24.0 * 365.0))

def get_age(ti, empty=None):
    if ti is None:
        return empty
    return int(float(ti.days) / 365.0 
                         + ti.seconds / (60.0 * 24.0 * 365.0))


def human_timedelta(td, empty=None):
    if td is None:
        return empty
    out = []
    if -td == abs(td):
        out.append("-")
        td = -td
    days = td.days
    if days == 1:
        out.append("1 Tag,")
    if days > 1:
        out.append("%d Tage," % days)

    secs = td.seconds
    hours = secs / 3600
    secs -= hours * 3600
    mins = secs / 60
    secs -= mins * 60

    out.append("%d:%02d:%02d" % (hours, mins, secs))

    return " ".join(out)


# formatters for the format_object funtion.
format_object_handlers = [
    (datetime.date, format_date),
    (datetime.datetime, format_datetime),
    (datetime.time, format_time),
    (type(None), lambda obj: ''),
    ]

def format_object(obj):
    """Return a string representation of obj suitable for the user.
    Date/time objects are formatted with the appropriate format_date
    function.  None objects are converted to an empty string.  Objects
    that are already strings (bytestrings or unicode) are returned as
    is.  Other objects are converted using str().
    """
    for typeobj, handler in format_object_handlers:
        if isinstance(obj, typeobj):
            return handler(obj)
    if isinstance(obj, basestring):
        return obj
    return str(obj)


# Fallback functions for the localized format_* functions. If localisation
# fails because of dates before 1900 (which aren't meaningfull anyway') the
# fallback functions are called. These functions will convert dates and time in
# german date and time formats

def dd_mm_YYYY(d, empty=''):
    """Format the date of the date/time object d as 'dd.mm.YYYY'.
    The parameter d may be a string or a datetime instance.  If d is
    None, the function return the optional parameter empty (default "").
    """
    # datetime.strftime() dislikes years before 1900.
    if d is None:
        return empty
    # TODO: what should happen if d is a unicode object?
    if type(d) == str:
        try:
            y, m, d = d.split('-')
            d = datetime(y, m, d)
        except:
            return empty
    return "%02d.%02d.%d" % (d.day, d.month, d.year)

def HH_MM(d):
    """Format the time object d as 'HH:MM'.
    The parameter d may be a time, datetime or date instance.  If it is
    a date instance, time 00:00 is assumed.  If d is None, the function
    returns the empty string.
    """
    # datetime.strftime() dislikes years before 1900.
    if d is None:
        return ''
    if type(d) is datetime.date:
        # not so smart
        d = datetime.datetime(d.year, d.month, d.day)
    return "%02d:%02d" % (d.hour, d.minute)

def dd_mm_yyyy_HH_MM(d, empty=''):
    """Format the time/date object d as 'dd.mm.YYYY HH:MM'.
    The parameter d may be a time, datetime or date instance.  If it is
    a date instance, time 00:00 is assumed.  If d is None, the function
    return the optional parameter empty (default"").
    """
    if d is None:
        return empty
    return dd_mm_YYYY(d) + " " + HH_MM(d)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
