# -*- coding: utf-8 -*-
import logging
import sys
import re
import functools

import psycopg2.extras

from pylons import app_globals as g

from mpulsweb.lib.db import db
from mpulsweb.lib.security import hasRole
from mpulsweb.model.phase import PhaseFactory, get_phase_successors


SAVE_SEARCH = re.compile(r'[^\w:;\-\. ]', re.UNICODE)


log = logging.getLogger(__name__)


class Search(object):

    VALID_OPTIONS = ("search_string",)
    MULTI_OPTIONS = ()

    SEARCH_QUERY = """
    SELECT %(fields)s
    FROM master_tbl_view m
    JOIN ka_status_tbl_view st ON m.id = st.master_id
    JOIN ka_benutzer_tbl b ON m.bearbeiter_id = b.id
    WHERE %(where)s
    %(order)s
    """

    SEARCH_FIELDS = """m.id, m.phase, %s st.status, st.zugriff, st.id as s_id,
    ARRAY(SELECT tag_id FROM nm_tags_cases_view WHERE case_id = m.id) AS tags
    """

    SEARCH_ORDER = """ORDER BY %s %s"""


    def __init__(self):
        pass

    @classmethod
    def supports_option(self, option):
        return option in self.VALID_OPTIONS

    def _parse_search_str(self, s):
        # TODO This is a very very simple approach. We should defentiley
        # implement some searchengine here
        search = []
        options = {}

        if s:
            s = SAVE_SEARCH.sub(u'', s)
            option_list = [o.strip() for o in s.split(';')]
            for option in option_list:
                if option.find(":") > 0:
                    try:
                        key, value = [x.strip() for x in option.split(":")]
                        if key in self.VALID_OPTIONS:
                            if key in self.MULTI_OPTIONS:
                                options.setdefault(key, []).append(int(value))
                            else:
                                options[key] = value
                    except ValueError, err:
                        print >> sys.stderr, \
                              "Found invalid search syntax in options: %s" % err
                else:
                    search.append(option)
        options["search_string"] = search
        return options

    def check_list(self, option, values, expected_type, convert=None):
        """Check that all items in values have the type expected_type.
        The return value is a list with all items of values that have
        the expected_type.  Values of other types are omitted from the
        result, but warnings are written to the log when that happens.
        If the optional parameter convert is given, it should be a
        function or other callable object and is applied to all items in
        the return value.
        """
        checked = []
        for v in values:
            if not isinstance(v, expected_type):
                log.warning("Omitting one of the values for option %r because"
                            " it does not have type %r."
                            " Value is %r, type of value is %r",
                            option, expected_type, v, type(v))
                continue
            if convert is not None:
                v = convert(v)
            checked.append(v)
        return checked

    def check_search_options(self, options):
        """Check and sanitize the search options.

        The return value is a dictionary with the options that are
        considered valid.  With string values converted to only contain
        characters as determined by the SAVE_SEARCH regular expression.
        """
        sanitize_string = functools.partial(SAVE_SEARCH.sub, u'')
        checked = dict()
        for key, value in options.iteritems():
            if key not in self.VALID_OPTIONS:
                log.warning("Omitting option %r because the key is not valid",
                            (key, value))
                continue
            if key in self.MULTI_OPTIONS:
                value = self.check_list(key, value, int)
            elif key == "search_string":
                value = self.check_list(key, value, basestring,
                                        sanitize_string)
            elif isinstance(value, basestring):
                value = sanitize_string(value)
            elif isinstance(value, int):
                # Nothing to do for ints.
                pass
            else:
                log.warning("Omitting option %r because the value has the"
                            " unsupported type %r", (key, value), type(value))
                continue

            # else value is most likely an int
            checked[key] = value
        return checked

    def _build_sql(self, options):
        pass

    def queryDB(self, sql):
        '''Does the search on the DB and returns the result of the search'''
        rows = []
        con, cur = None, None
        try:
            con = db.getConnection()
            cur = con.cursor(cursor_factory=psycopg2.extras.DictCursor)
            cur.execute(sql)
            log.debug('QueryDB: %s' % cur.query)
            rows = cur.fetchall()
        finally:
            db.recycleConnection(con, cur)
        return rows

    def perform(self, search):
        """Returns the result set of a search.
        The parameter search should be a dictionary with the search
        options.  Alternatively, for backwards compatibility it may be a
        string containing the search options in the format parsed by the
        _parse_search_str method.
        """
        if isinstance(search, basestring):
            options = self._parse_search_str(search)
        else:
            options = self.check_search_options(search)
        return self.queryDB(self._build_sql(options))


class CaseSearch(Search):

    VALID_OPTIONS = (Search.VALID_OPTIONS
                     + ("state", "phase", "own", "standin", "editor", "branch",
                        "sdate", "edate", "sort_field", "sort_order"))
    MULTI_OPTIONS = Search.MULTI_OPTIONS + ('state', 'phase')


    def __init__(self, retrieve_fields, match_fields):
        Search.__init__(self)
        self.retrieve_fields = retrieve_fields
        self.match_fields = match_fields

    def get_fields(self, options):
        retrieve = ", ".join("m." + field for field in self.retrieve_fields)
        if retrieve:
            retrieve = "%s," % retrieve
        return self.SEARCH_FIELDS % retrieve

    def validate_form_params(self, params):
        # FIXME: We import SearchCaseForm here because importing it
        # during import doesn't work.  The validators module performs
        # translations and accesses app_globals during import, so it has
        # to be imported when a HTTP-request is being handled.  The
        # search module is imported early on when the application starts
        # so validators cannot be imported then.
        from mpulsweb.lib.validators import SearchCaseForm
        return SearchCaseForm().to_python(params)

    def convert_form_parameters(self, options, user, extended):
        """Convert form parameters to search options.
        The parameter user is the user object for the current user.
        If extended is true, the search is an extended search, otherwise
        it's the more limited default search.  In the limited default
        search, some options are not included in the return value.
        """
        search_options = dict()

        # Searchstring
        if options.get('search_str'):
            search_options["search_string"] = [options.get('search_str')]

        # Status
        state = options.get("state")
        if state:
            # for backwards compatibility we always include -1, but only
            # if the user actually wants to search based on the state,
            # i.e. when at least one state value was given.
            search_options["state"] = [-1] + list(state)

        # Responsibilty
        # shown cases where the user is editor, or standin or both
        if options.get("own"):
            search_options["own"] = user.id
        if options.get("standin"):
            search_options["standin"] = user.id

        # Editor
        # shown cases where the user is editor or all (-1)
        if options.get("editor") and options.get("editor") != -1:
            search_options["editor"] = options.get("editor")

        if extended:
            # Phase
            if "phase" in options:
                # Get phasepart ids for the selected phasepairs. Only if
                # pairs are defined.
                pairs = g.mpuls_config.get("phases", "pairs")
                if pairs:
                    phases = []
                    for pp in options["phase"]:
                        phases.extend((int(p) for p in pairs[0].get(str(pp),
                                                                    [-1])))
                    if phases:
                        search_options["phase"] = phases

            # Dates
            if (options.get("sdate") is not None
                and options.get("edate") is not None):
                search_options["sdate"] = options.get("sdate")
                search_options["edate"] = options.get("edate")

            # gender
            if options.get("gender", -2) != -2:
                search_options["gender"] = options["gender"]

            # branch office
            if options.get("branch"):
                search_options["branch"] = options["branch"]

        return search_options

    def generic_int_choice_clause(self, options, key, column):
        """Return a search condition for an option with multiple int values.

        The allowed values are the values in the list of ints stored in
        the options dictionary under key.  There are three cases to
        consider:

          1. options does not contain key at all.  This means the value
             doesn't matter and thus the method returns 'TRUE'.

          2. the values list is empty.  This means no value matches and
             no case will match.  The method returns 'FALSE'

          3. The value is a list.  In this case the method returns a
             condition that selects cases where the column given by the
             string column has one of the values in the list.
        """
        if key in options:
            values = options[key]
            if values:
                return ('%s IN (%s)'
                        % (column, ','.join([str(v) for v in values])))
            return "FALSE"
        return "TRUE"

    def get_allowed_viewer_clause(self, options):
        """Return the search condition that selects based on owner or standin.
        This means the search condition a case manager user can use to
        limit the search to cases that the user can see or modify due to
        the user being the cases's owner or a standin.

        If the current user is an admin this method should return
        'TRUE'.

        Derived classes may override this method if necessary.
        """
        if hasRole(['admin_ka', 'pb_ka']):
            return "TRUE"

        which_user = 0
        allowed_viewer = "TRUE"

        if options.has_key('own'):
            which_user += 2
        if options.has_key('standin'):
            which_user += 1

        if which_user == 1:
            allowed_viewer = "m.bearbeiter_id <> %s" % options.get('standin')
        if which_user == 2:
            allowed_viewer = "m.bearbeiter_id = %s" % options.get('own')
        if which_user == 3:
            allowed_viewer = "TRUE"

        return allowed_viewer

    def get_editor_clause(self, options):
        """Return the search condition to select cases with a specific owner.
        This is the condition an admin user can use to limit the search
        to cases of a specific case manager.

        Derived classes may override this method if necessary.
        """
        if options.has_key('editor'):
            return "m.bearbeiter_id = %s" % options['editor']
        return "TRUE"

    def get_branch_clause(self, options):
        """Return the search condition to select cases based on branch office.
        Derived classes may override this method if necessary.
        """
        branch = options.get('branch')
        if branch is not None:
            return "b.filiale = '%s'" % branch
        return "TRUE"

    def get_needle_clause(self, options):
        """Return the search condition to select cases based on search terms.
        Derived classes may override this method if necessary.
        """
        search = options.get("search_string", ())
        if search:
            needle_expr = ("((%s) OR (%s))"
                           % (" OR ".join("m.%s ~* '%%(needle)s'"
                                          % field
                                          for field in self.match_fields),
                              " AND ".join("m.%s IS NULL"
                                           % field
                                           for field in self.match_fields)))
            return needle_expr % dict(needle=" ".join(search))
        return "TRUE"

    def get_phases_dates_clause(self, options):
        """Return the search condition selecing cases based on phase and dates.

        The phases used for the return value are given as the value of
        the 'phase' key of the options dictionary as a list of phase
        ids.  If the options dictionary does not contain the 'phase'
        key, the value is considered to equivalent to a list containing
        all possible phase ids.

        The start and end dates are given as the values of the 'sdate'
        and 'edate' keys of the options dictionary.

        If both a start date and an end date are given, the returned
        clause matches all cases which were in one of the given phases
        during the given time.  More precisely, it matches all cases
        where the date interval of one of the phases in the phases list
        overlaps the interval specified for the search and the current
        phase is that phase or one the successor phases of that phase.

        Otherwise, the clause matches all cases that are currently in
        one of the given phases.

        Note that the phases are specified as a list.  If the options
        dictionary contains the key 'phase' but the list of phases it
        maps to is empty, the claus will be equivalent to 'FALSE' and no
        case will match.

        Derived classes may override this method if necessary.
        """
        sd = options.get('sdate')
        ed = options.get('edate')

        if sd is None or ed is None:
            # no date range given -> match according to current phase
            return self.generic_int_choice_clause(options, "phase", "m.phase")

        phases = options.get('phase')
        phase_clauses = []
        phase_list = PhaseFactory().build()
        for p in phase_list:
            sp = p.getStart()
            ep = p.getEnd()

            if sp.datefield is None or ep.datefield is None:
                # ignore phases without datefields.  This is usually
                # only the 'unknown' phase.
                continue

            # if phases is None, it means the phases parameter was
            # omitted altogether and all phases should match.
            if phases is None or int(sp.id) in phases or int(ep.id) in phases:
                # a phase can either be finished, in which case we have
                # both start and end dates for the phase, or it may be
                # still running, in which case we have a start date but
                # no end date yet.  This means that cases where the
                # start date of the phase has not been set yet should
                # not match.  However, cases where the end date of the
                # phase has not been set yet should match.
                successors = get_phase_successors(sp.phase, phase_list)
                phase_clauses.append("(%s IS NOT NULL"
                                     " AND (%s::date <= '%s'::date)"
                                     " AND "
                                     "(coalesce(%s, now())::date "
                                     ">= '%s'::date)"
                                     " AND phase IN (%s))"
                                     % (sp.datefield, sp.datefield, ed,
                                        ep.datefield, sd,
                                        ",".join(str(s) for s in successors)))
        if phase_clauses:
            return "(%s)" % " OR ".join(phase_clauses)
        return "FALSE"

    def get_status_clause(self, options):
        """Return the search condition to select cases based on status.
        Derived classes may override this method if necessary.
        """
        return self.generic_int_choice_clause(options, "state", "st.status")

    def get_gender_clause(self, options):
        """Return the search condition to select cases based on gender.
        The default implementation simply returns 'TRUE'.

        Derived classes should override this method and include 'gender'
        in their VALID_OPTIONS value.
        """
        return "TRUE"

    def get_where_subclauses(self, options):
        """Return a list with the 'where' sub-clauses as strings.
        This base method calls the get_*_clause methods to gather the
        commonly used subclauses.  Derived classes should extend this
        method and add new strings to the return value if they need more
        clauses.
        """
        return [
            self.get_needle_clause(options),
            self.get_status_clause(options),
            self.get_allowed_viewer_clause(options),
            self.get_editor_clause(options),
            self.get_branch_clause(options),
            self.get_phases_dates_clause(options),
            self.get_gender_clause(options),
            ]

    def get_where(self, options):
        """Return the 'where' clause for the search.
        The clause contains all clauses returned by self.get_where_subclauses
        combined by AND.
        """
        return " AND ".join(self.get_where_subclauses(options))

    def get_order(self, options):
        sort_field = options.get('sort_field', "m.id")
        sort_order = options.get('sort_order', "desc")
        return self.SEARCH_ORDER % (sort_field, sort_order)

    def _build_sql(self, options):
        sql_fields = {}
        sql_fields['fields'] = self.get_fields(options)
        sql_fields['where'] = self.get_where(options)
        sql_fields['order'] = self.get_order(options)

        SQL = self.SEARCH_QUERY % sql_fields
        return SQL

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
