# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
#

import logging
from datetime import datetime, timedelta

import mpulsweb.lib.helpers as h
from mpulsweb.lib.base import session, g
from mpulsweb.lib.db import db
from mpulsweb.model.case import MpulsCaseOverview
from mpulsweb.model.agencysettings import Agency
from mpulsweb.controllers.case_overview import default_phases 


GET_GLOBAL_APPOINTMENT_SQL = """\
SELECT id, master_id, sdatum, edatum, name, beschreibung, art
FROM ka_global_termine_tbl_view
WHERE id = %(id)s
"""
GET_GLOBAL_APPOINTMENTLIST_SQL = """\
SELECT id, master_id, sdatum, edatum, name, beschreibung, art
FROM ka_global_termine_tbl_view
WHERE coalesce(edatum, sdatum)::date+1 >= %(date)s::date ORDER BY sdatum
"""
DELETE_GLOBAL_APPOINTMENT_SQL = "SELECT delete_global_termin_ds(%(id)s)"
CREATE_GLOBAL_APPOINTMENT_SQL = "SELECT create_global_termin_ds()"
STORE_GLOBAL_APPOINTMENT_SQL = """\
UPDATE ka_global_termine_tbl_view
SET
    sdatum = %(start_date)s::timestamp,
    edatum = %(end_date)s::timestamp,
    name = %(title)s,
    beschreibung = %(description)s,
    art = %(type)s
WHERE id = %(id)s
"""

GET_CASE_APPOINTMENT_SQL = """\
SELECT id, master_id, sdatum, edatum, name, beschreibung, art
FROM ka_fall_termine_tbl_view
WHERE id = %(id)s
"""

GET_CASE_APPOINTMENTLIST_SQL = """\
SELECT id, master_id, sdatum, edatum, name, beschreibung, art
FROM ka_fall_termine_tbl_view
WHERE master_id = %(id)s
      AND coalesce(edatum, sdatum)::date+1 >= %(date)s::date
      AND art = 0
ORDER BY sdatum
"""

GET_CASE_REMINDERLIST_SQL = """\
SELECT id, master_id, sdatum, edatum, name, beschreibung, art
FROM ka_fall_termine_tbl_view
WHERE master_id = %(id)s AND art = 1
ORDER BY sdatum
"""

GET_ALLCASES_APPOINTMENTLIST_SQL = """\
SELECT a.id, a.master_id, a.sdatum, a.edatum, a.name, a.beschreibung, a.art
FROM ka_fall_termine_tbl_view a 
JOIN master_tbl_view m ON m.id = a.master_id 
JOIN ka_status_tbl_view s ON m.id = s.master_id
WHERE coalesce(edatum, sdatum)::date+1 >= %(date)s::date
      AND art = 0 AND m.bearbeiter_id = %(editor)s AND s.status = 1
ORDER BY sdatum
"""

GET_ALLCASES_REMINDERLIST_SQL = """\
SELECT a.id, a.master_id, a.sdatum, a.edatum, a.name, a.beschreibung, a.art
FROM ka_fall_termine_tbl_view a 
JOIN master_tbl_view m ON m.id = a.master_id
JOIN ka_status_tbl_view s ON m.id = s.master_id
WHERE art = 1 AND m.bearbeiter_id = %(editor)s AND s.status = 1
ORDER BY sdatum
"""

DELETE_CASE_APPOINTMENT_SQL = "SELECT delete_fall_termin_ds(%(id)s)"
CREATE_CASE_APPOINTMENT_SQL = "SELECT create_fall_termin_ds(%(id)s)"
STORE_CASE_APPOINTMENT_SQL = """\
UPDATE ka_fall_termine_tbl_view
SET
    sdatum = %(start_date)s::timestamp,
    edatum = %(end_date)s::timestamp,
    name = %(title)s,
    beschreibung = %(description)s,
    art = %(type)s
WHERE id = %(id)s
"""

GET_OLDACTIVE_CASES_4_ADMIN = """
SELECT m.id, s.zugriff, getLastCaseAppointment(m.id) AS last_date
FROM
  (SELECT * FROM ka_status_tbl_view
   WHERE zugriff < now() - interval '1 day' * %%(maxdays)s AND status <> 5) s
   JOIN master_tbl_view m ON m.id = s.master_id
WHERE phase in (%s)
AND (%s)
"""

GET_OLDACTIVE_CASES = """
SELECT m.id, s.zugriff, getLastCaseAppointment(m.id) AS last_date
FROM
  (SELECT * FROM ka_status_tbl_view
   WHERE zugriff < now() - interval '1 day' * %%(maxdays)s AND status <> 5) s
   JOIN master_tbl_view m ON m.id = s.master_id
WHERE phase in (%s) AND (%s) AND m.bearbeiter_id = %%(editor)s
"""

GET_OLDFINISHED_CASES_4_ADMIN = """\
SELECT m.id, s.zugriff, getLastCaseAppointment(m.id) as last_date
FROM master_tbl_view m JOIN ka_status_tbl_view s ON s.master_id = m.id
WHERE s.status <> 5
      AND extract(day from now()-s.zugriff) > %%(maxdays)s
      AND phase in (%s)
      AND (%s)"""

GET_OLDFINISHED_CASES = """\
SELECT m.id, s.zugriff, getLastCaseAppointment(m.id) as last_date
FROM master_tbl_view m JOIN ka_status_tbl_view s ON s.master_id = m.id
WHERE s.status <> 5
      AND m.bearbeiter_id = %%(editor)s
      AND extract(day from now()-s.zugriff) > %%(maxdays)s
      AND phase in (%s)
      AND (%s)
"""

log = logging.getLogger(__name__)


class AppointmentFactory:

    def loadById(self, id):
        appointment = None
        try:
            conn = db.getConnection()
            cur = conn.cursor()
            cur.execute(self.get_sql, {'id': id})
            row = cur.fetchone()
            appointment = self.createByDBRow(row)
        finally:
            db.recycleConnection(conn, cur)
        return appointment

    def createByData(self, id=None, title=None, sdate=None, edate=None,
                     desc=None, case_id=None, type=0):
        appointment = self.constructor()
        appointment.id = id
        appointment.start_date = sdate
        appointment.end_date = edate
        appointment.title = title
        appointment.description = desc
        appointment.case_id = case_id
        appointment.type = type
        return appointment

    def createByDBRow(self, row):
        id = row[0]
        master_id = row[1]
        sdate = row[2]
        edate = row[3]
        name = h.ensure_unicode(row[4])
        desc = h.ensure_unicode(row[5])
        type = row[6]
        return self.createByData(id, name, sdate, edate, desc, master_id, type)

    def createNew(self, case_id=None):
        try:
            conn = db.getConnection()
            cur = conn.cursor()
            if case_id:
                cur.execute(self.create_sql, {'id': case_id})
            else:
                cur.execute(self.create_sql)
            row = cur.fetchone()
            id = row[0]
            return self.createByData(id)
        finally:
            db.recycleConnection(conn, cur)
        return None


class CaseAppointmentFactory(AppointmentFactory):

    def __init__(self):
        self.get_sql = GET_CASE_APPOINTMENT_SQL
        self.create_sql = CREATE_CASE_APPOINTMENT_SQL
        self.constructor = CaseAppointment


class CaseReminderFactory(AppointmentFactory):

    def __init__(self):
        self.get_sql = GET_CASE_APPOINTMENT_SQL
        self.create_sql = CREATE_CASE_APPOINTMENT_SQL
        self.constructor = CaseReminder


class GlobalAppointmentFactory(AppointmentFactory):

    def __init__(self):
        self.get_sql = GET_GLOBAL_APPOINTMENT_SQL
        self.create_sql = CREATE_GLOBAL_APPOINTMENT_SQL
        self.constructor = GlobalAppointment


class AppointmentOverview:

    def _load(self, factory, sql, values):
        self.appointment_list = []
        conn, cur = None, None
        try:
            conn = db.getConnection()
            cur = conn.cursor()
            cur.execute(sql, values)
            rows = cur.fetchall()
            for row in rows:
                self.appointment_list.append(factory.createByDBRow(row))
        finally:
            db.recycleConnection(conn, cur)

    def getDatasets(self):
        return self.appointment_list

    def setDatasets(self, list):
        self.appointment_list = list

    def join(self, overview):
        self.appointment_list.extend(overview.getDatasets())

    def sort(self, cmp=None):
        self.appointment_list = sorted(self.appointment_list)


class CaseAppointmentOverview(AppointmentOverview):

    def __init__(self, caseid=None):
        afactory = CaseAppointmentFactory()
        today = datetime.now().replace(hour=0, minute=0, second=0,
                                       microsecond=0)
        if caseid:
            values = {'id': caseid, 'date': today}
            sql = GET_CASE_APPOINTMENTLIST_SQL
        else:
            user = session['USER_AUTHORIZED']
            values = {'date': today, 'editor': user.id}
            sql = GET_ALLCASES_APPOINTMENTLIST_SQL
        self._load(afactory, sql, values)


class GlobalAppointmentOverview(AppointmentOverview):

    def __init__(self):
        afactory = GlobalAppointmentFactory()
        today = datetime.now().replace(hour=0, minute=0, second=0,
                                       microsecond=0)
        values = {'date': today}
        sql = GET_GLOBAL_APPOINTMENTLIST_SQL
        self._load(afactory, sql, values)


class CaseReminderOverview(AppointmentOverview):

    def __init__(self, caseid=None):
        afactory = CaseReminderFactory()
        today = datetime.now().replace(hour=0, minute=0, second=0,
                                       microsecond=0)
        if caseid:
            values = {'id': caseid, 'date': today}
            sql = GET_CASE_REMINDERLIST_SQL
        else:
            user = session['USER_AUTHORIZED']
            values = {'date': today, 'editor': user.id}
            sql = GET_ALLCASES_REMINDERLIST_SQL
        self._load(afactory, sql, values)


class Appointment:

    def __init__(self):
        self.id = None
        self.title = None
        self.start_date = None
        self.end_date = None
        self.description = None
        self.case_id = None
        # available types
        # 0: normal appointment
        # 1: Reminder
        self.type = 0

    def __cmp__(self, other):
        return cmp(self.start_date, other.start_date)

    def _convertStr2Date(self, str, format=None):
        str = str.strip()
        date, time = str.split(' ')
        try:
            d, m, Y = date.split('.')
        except:
            return None
        try:
            H, M = time.split(':')
        except:
            H, M = (0, 0)
        return datetime(int(Y), int(m), int(d), int(H), int(M))

    def _convert2Datetime(self, mdate, mtime):
        log.debug("%s, %s" % (mtime, type(mtime)))
        try:

            try:
                y = mdate.year
                m = mdate.month
                d = mdate.day
            except:
                return None

            # If no time id provided set to 00:00:00
            try:
                H = mtime.hour
                M = mtime.minute
                S = mtime.second
            except:
                H = 0
                M = 0
                S = 0
            return datetime(y, m, d, H, M, S)

        except Exception, e:
            log.error(e)
            return None

    def getStartDate(self):
        return self.start_date

    def getEndDate(self):
        return self.end_date

    def getDuration(self, empty=''):
        if self.end_date:
            return h.human_timedelta(self.end_date - self.start_date)
        return empty

    def setData(self, data):
        """Sets values of the appointment to data.
        Data is a dict returned from formencode
        """
        self.title = data.get('title', self.title)
        self.type = data.get('type', self.type)

        # Set date
        self.start_date = self._convert2Datetime(data.get('start_date'),
                                                 data.get('start_time'))

        end_date = self._convert2Datetime(data.get('end_date'),
                                          data.get('end_time'))

        # If no enddate is given assume a 24h appointment
        if not end_date:
            # Add 24 hours to startdate
            self.end_date = self.start_date + timedelta(hours=24)
        else:
            self.end_date = end_date

        self.description = data.get('description', self.description)
        self.case_id = data.get('case_id', self.case_id)

    def store(self):
        try:
            fields = {'id': self.id, 'case_id': self.case_id, 'type': self.type,
                      'title': self.title, 'description': self.description,
                      'start_date': self.start_date, 'end_date': self.end_date,
                      }
            db.execute(self.store_sql, fields)
            return True
        except Exception:
            log.exception("Error while storing appointment %r for case %r"
                          % (self.id, self.case_id))
        return False

    def isGlobal(self):
        return self.case_id is None

    def isReminder(self):
        return self.type == 1

    def delete(self):
        db.execute(self.delete_sql, {'id': self.id})
        return True


class GlobalAppointment(Appointment):

    def __init__(self):
        Appointment.__init__(self)
        self.type = 0
        self.store_sql = STORE_GLOBAL_APPOINTMENT_SQL
        self.delete_sql = DELETE_GLOBAL_APPOINTMENT_SQL


class CaseAppointment(Appointment):

    def __init__(self):
        Appointment.__init__(self)
        self.type = 0
        self.store_sql = STORE_CASE_APPOINTMENT_SQL
        self.delete_sql = DELETE_CASE_APPOINTMENT_SQL


class CaseReminder(CaseAppointment):

    def __init__(self):
        CaseAppointment.__init__(self)
        self.type = 1

# MAXSAVETIME

class MaxSaveTimeReminderOverview(AppointmentOverview):

    def __init__(self):
        self.appointment_list = []
        agency = Agency()
        maxage = agency.getMaxSavetime()
        cases = MpulsCaseOverview()
        user = session['USER_AUTHORIZED']

        # 0. Load ids and some additional info on cases which either are active
        # but seems to be orphaned or are finished and not edited for
        # longer than maxage.
        # orphaned means that all last appointments, aids or the
        # modification date is older than maxage
        oldAndActive = self.getOldAndActiveCases(maxage, user)
        oldAndFinished = self.getOldAndFinishedCases(maxage, user)

        # 1. Load the user's cases so we can check the maximum savetime on them
        default_search_options = ['own:%s' % user.id, 'standin:%s' % user.id,
                                  'state:1', 'state:2', 'state:3']
        # Get phasepart ids for the selected phasepairs. Only if pairs are
        # defined.
        for p in default_phases():
            default_search_options.append('phase:%s' % p)
        cases.search(";".join(default_search_options))

        # 2. filter cases which seem to be orphaned and create reminders
        oldAndActiveSet = set(oldAndActive)
        old_cases = [case for case in cases.getDatasets()
                     if case.id in oldAndActiveSet]
        log.debug('oldcases: %s' % old_cases)
        for case in old_cases:
            info = oldAndActive[case.id]
            last = info['last_date']
            appointment = self.makeAppointment(case, last, maxage,
                                                 finished=False)
            self.appointment_list.append(appointment)


        # 3. filter cases which are finished and not edited and create reminders
        oldAndFinishedSet = set(oldAndFinished)
        old_cases = [case for case in cases.getDatasets()
                     if case.id in oldAndFinishedSet]
        for case in old_cases:
            info = oldAndFinished[case.id]
            access = info['access']
            appointment = self.makeAppointment(case, access, maxage,
                                                 finished=True)
            self.appointment_list.append(appointment)

    def getOldAndFinishedCases(self, days, user):
        conn, cur = None, None
        list = {}
        remindercondition = g.mpuls_config.get('case', 'remindercondition') 
        try:
            conn = db.getConnection()
            cur = conn.cursor()
            if user.isAdmin():
                sql = (GET_OLDFINISHED_CASES_4_ADMIN
                       % (",".join(g.mpuls_config.get('phases', 'finished')), remindercondition))
                cur.execute(sql, {'maxdays': days})
            else:
                sql = (GET_OLDFINISHED_CASES
                       % (",".join(g.mpuls_config.get('phases', 'finished')), remindercondition))
                cur.execute(sql, {'maxdays': days, 
                                  'editor': user.id})
            result = cur.fetchall()
            for r in result:
                list[r[0]] = {'access':r[1], 'last_date': r[2]}
            return list
        finally:
            db.recycleConnection(conn, cur)
        return False

    def getOldAndActiveCases(self, days, user):
        conn, cur = None, None
        list = {}
        remindercondition = g.mpuls_config.get('case', 'remindercondition') 
        try:
            conn = db.getConnection()
            cur = conn.cursor()
            if user.isAdmin():
                sql = (GET_OLDACTIVE_CASES_4_ADMIN
                       % (",".join(g.mpuls_config.get('phases', 'active')), remindercondition))
                cur.execute(sql, {'maxdays': days, 
                                  'editor': user.id})
            else:
                sql = (GET_OLDACTIVE_CASES
                       % (",".join(g.mpuls_config.get('phases', 'active')), remindercondition))
                cur.execute(sql, {'maxdays': days, 
                                  'editor': user.id})
            result = cur.fetchall()
            for r in result:
                list[r[0]] = {'access':r[1], 'last_date': r[2]}
            return list
        finally:
            db.recycleConnection(conn, cur)
        return False

    def makeAppointment(self, case, last_date, maxage, finished):
        factory = CaseAppointmentFactory()
        type = 1
        title = u"%s, %s" % (case.last_name, case.first_name)
        age = datetime.now() - last_date
        mydate = datetime.now() - timedelta(age.days - int(maxage))
        if finished:
            desc = (u"Achtung! Die maximale Speicherdauer von %s Tagen"
                    u" ist seit %s Tagen für diese Fallakte überschritten!"
                    % (maxage, age.days-int(maxage)))
        else:
            desc = (u"Inaktive Fallakte? Bitte prüfen Sie,"
                    u" ob der Fall noch bearbeitet wird.")
        appointment = factory.createByData(None, title, mydate, None, desc,
                                           case.id, type)
        return appointment

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf-8:
