BEGIN;
SET ROLE :db_adm;

-- ka_benutzergruppe_tbl
-- Tabelle fuer die Benutzergruppen. Jeder Benutzer erhaelt eine
-- Benutzergruppe in der er exklusiv Mitglied ist. 
-- Weitere Benutzergruppen können angelegt werden, die weitere Nutzer
-- enthalten können

CREATE TABLE ka_benutzergruppe_tbl
(
	id serial NOT NULL,
	name varchar(128) NOT NULL,
	CONSTRAINT ka_benutzergruppe_tbl_pkey PRIMARY KEY (id)
);

ALTER TABLE ka_benutzer_tbl
ADD COLUMN gid int4 REFERENCES ka_benutzergruppe_tbl(id) ON DELETE CASCADE;

-- ka_nm_benutzer_gruppe_tbl
-- Tabelle zur Zuordnung zwischen Benutzern und Benutzergruppen.
CREATE TABLE nm_benutzer_gruppe_tbl
(
	id serial NOT NULL,
	gruppe_id int4 REFERENCES ka_benutzergruppe_tbl(id) ON DELETE CASCADE,
	benutzer_id int4 REFERENCES ka_benutzer_tbl(id) ON DELETE CASCADE,
	CONSTRAINT nm_benutzer_gruppe_tbl_pkey PRIMARY KEY (id)
);

-- Funktionen zu Erstellen und Loeschen neuer Eintraeger in der
-- Benutzergruppen Tabelle

CREATE OR REPLACE FUNCTION create_benutzergruppe_ds(gname varchar) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    INSERT INTO ka_benutzergruppe_tbl (name) VALUES (gname);
    RETURN currval('ka_benutzergruppe_tbl_id_seq');
END;
$$ ;

REVOKE ALL ON FUNCTION create_benutzergruppe_ds( varchar ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_benutzergruppe_ds( varchar ) TO GROUP :admin_ka_group;


CREATE OR REPLACE FUNCTION delete_benutzergruppe_ds(gid integer) RETURNS void 
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
BEGIN
    DELETE FROM ka_benutzergruppe_tbl WHERE id = gid;
END;
$$ ;

REVOKE ALL ON FUNCTION delete_benutzergruppe_ds( integer ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_benutzergruppe_ds( integer ) TO GROUP :admin_ka_group;

-- Sichten auf die Benutzergruppe

CREATE OR REPLACE VIEW ka_benutzergruppe_tbl_view AS 
SELECT g.id, coalesce(b.nachname ||', '|| b.vorname, g.name) AS name FROM ka_benutzer_tbl b RIGHT JOIN ka_benutzergruppe_tbl g ON g.id = b.gid;

CREATE OR REPLACE RULE ka_benutzergruppe_delete_rule AS ON DELETE TO 
ka_benutzergruppe_tbl_view DO INSTEAD NOTHING;

CREATE OR REPLACE RULE ka_benutzergruppe_insert_rule AS ON INSERT TO 
ka_benutzergruppe_tbl_view DO INSTEAD NOTHING;

CREATE OR REPLACE RULE ka_benutzergruppe_update_rule AS ON UPDATE TO ka_benutzergruppe_tbl_view 
	DO INSTEAD  UPDATE ka_benutzergruppe_tbl SET 
	name = new.name 
	WHERE ka_benutzergruppe_tbl.id = new.id;

GRANT SELECT ON ka_benutzergruppe_tbl_view TO GROUP :cm_ka_group;
GRANT SELECT ON ka_benutzergruppe_tbl_view TO GROUP :admin_ka_group;
GRANT UPDATE ON ka_benutzergruppe_tbl_view TO GROUP :admin_ka_group;

-- Funktionen zu Erstellen und Loeschen neuer Eintraeger in der
-- nm_benutzer_gruppe Tabelle

CREATE OR REPLACE FUNCTION create_nm_benutzer_gruppe_ds(bid integer, gid integer) RETURNS integer
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    INSERT INTO nm_benutzer_gruppe_tbl (benutzer_id, gruppe_id) VALUES (bid, gid);
    RETURN currval('nm_benutzer_gruppe_tbl_id_seq');
END;
$$ ;
REVOKE ALL ON FUNCTION create_nm_benutzer_gruppe_ds( integer, integer ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION create_nm_benutzer_gruppe_ds( integer, integer ) TO GROUP :admin_ka_group;

CREATE OR REPLACE FUNCTION delete_nm_benutzer_gruppe_ds( bid integer, gid integer) RETURNS void 
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
BEGIN
    DELETE FROM nm_benutzer_gruppe_tbl WHERE gruppe_id = gid AND benutzer_id = bid;
END;
$$ ;
REVOKE ALL ON FUNCTION delete_nm_benutzer_gruppe_ds( integer, integer ) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION delete_nm_benutzer_gruppe_ds( integer, integer ) TO GROUP :admin_ka_group;

-- 
-- Sicht auf die Benutzer-Gruppen Transitionstabelle
--

CREATE OR REPLACE VIEW nm_benutzer_gruppe_tbl_view AS 
SELECT nm_benutzer_gruppe_tbl.id, nm_benutzer_gruppe_tbl.benutzer_id,
nm_benutzer_gruppe_tbl.gruppe_id 
FROM nm_benutzer_gruppe_tbl;

CREATE OR REPLACE RULE nm_benutzer_gruppe_delete_rule AS ON DELETE TO 
nm_benutzer_gruppe_tbl_view DO INSTEAD NOTHING;

CREATE OR REPLACE RULE nm_benutzer_gruppe_insert_rule AS ON INSERT TO
nm_benutzer_gruppe_tbl_view DO INSTEAD NOTHING;

CREATE OR REPLACE RULE nm_benutzer_gruppe_update_rule AS ON UPDATE TO
nm_benutzer_gruppe_tbl_view DO INSTEAD NOTHING;

GRANT SELECT ON nm_benutzer_gruppe_tbl_view TO GROUP :cm_ka_group;
GRANT SELECT ON nm_benutzer_gruppe_tbl_view TO GROUP :admin_ka_group;

-- Updatefuntktion zum Erstellen von Benutzergruppen für jeden Benutzer

CREATE OR REPLACE FUNCTION create_user_groups() RETURNS integer AS $$
DECLARE
	benutzer RECORD;
	gname varchar;
	tgid   integer;
	maxid integer;
BEGIN
    FOR benutzer IN SELECT * FROM ka_benutzer_tbl ORDER BY id LOOP
        gname := benutzer.login;
        tgid   := benutzer.id;
        INSERT INTO ka_benutzergruppe_tbl (id, name) VALUES (tgid, gname);
        INSERT INTO nm_benutzer_gruppe_tbl (gruppe_id, benutzer_id) VALUES (tgid, tgid);
        UPDATE ka_benutzer_tbl SET gid = tgid WHERE id = tgid;
    END LOOP;
    SELECT INTO maxid max(id) from ka_benutzergruppe_tbl;
    PERFORM setval('ka_benutzergruppe_tbl_id_seq', maxid);
    RETURN 1;
END;
$$ LANGUAGE plpgsql;

SELECT create_user_groups();
DROP FUNCTION create_user_groups();

-- Umbiegen des FK Constraint in ther benutzer_master tabelle, so das die
-- benutzer_id nun auf die benutzergruppe zeigt!
ALTER TABLE nm_benutzer_master_tbl DROP CONSTRAINT has_user_fk;
ALTER TABLE nm_benutzer_master_tbl ADD CONSTRAINT has_group_fk FOREIGN KEY (benutzer_id) REFERENCES ka_benutzergruppe_tbl (id) ON DELETE CASCADE; 

-- 03functions
--
-- Anpassung des bestehenden Codes
--
-- 01waskatmp
-- Siehe 02_vertretung_update_views.sql 

-- 02tables

-- ALTER TABLE ka_benutzer_tbl
-- ADD COLUMN gid int4 REFERENCES ka_benutzergruppe_tbl(id) ON DELETE CASCADE;

-- ------------------------------------------------------------------------
-- Anlegen einer Rolle in der DB einer KA:
-- - Ueber waska_create_role() Datenbank-Rolle anlegen und 
-- - dann Eintragungen in ka_benutzer_tbl machen.
-- - Benutzergruppe anlegen
-- - Benutzer in die Benutzergruppe eintragen
CREATE OR REPLACE FUNCTION ka_create_role (
        agency          varchar(52),
        grouprole       varchar(64),
        loginname       varchar(64),
        firstname       varchar(64),
        surname         varchar(64),
        phone           varchar(64),
        room            varchar(64),
        branchoffice    varchar(64),
        activated       boolean,
        setpassword     boolean
    ) RETURNS varchar(9) 
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
    newrole    RECORD;
    groupid    integer;
    userid     integer;
BEGIN
    SELECT INTO newrole * FROM waska_create_role(loginname, grouprole, agency)
        AS (db_name varchar(64), db_user varchar(64), pw varchar(9));
    EXECUTE 'GRANT CONNECT ON DATABASE ' || newrole.db_name 
                || ' TO ' || newrole.db_user; 
    INSERT INTO ka_benutzer_tbl ( login, vorname, nachname,
        telefon, raumnummer, filiale,
        aktiviert, passwort_aendern, gid) 
        VALUES
        (newrole.db_user, firstname, surname,
        phone, room, branchoffice,
        activated, setpassword, NULL);
    SELECT INTO userid currval('ka_benutzer_tbl_id_seq');
    SELECT INTO groupid create_benutzergruppe_ds(surname ||', '||firstname);
    UPDATE ka_benutzer_tbl SET gid = groupid WHERE id = userid;
    INSERT INTO nm_benutzer_gruppe_tbl (benutzer_id, gruppe_id) VALUES (userid, groupid);

    RETURN newrole.pw::varchar(9);
END;
$$ ;

CREATE OR REPLACE FUNCTION ka_delete_role (
        loginname       varchar(64),
        agency          varchar(52)
    ) RETURNS VOID 
    LANGUAGE 'plpgsql' VOLATILE EXTERNAL SECURITY DEFINER
    AS $$
DECLARE
    db_user  varchar(128);
    groupid   integer;
BEGIN
    db_user := quote_ident('ka_' || agency || '_' || loginname);
    SELECT INTO groupid gid FROM ka_benutzer_tbl WHERE login = db_user;
    DELETE FROM ka_benutzer_tbl WHERE login = db_user;
    DELETE FROM ka_benutzergruppe_tbl WHERE id = groupid;
    PERFORM waska_drop_role(loginname, agency);
END;
$$ ;

-- Die Funktionen zu erstellen und Löschen von Einträgen in der
-- nm_benutzer_master tabelle 
-- * create_master_ds
-- * create_nm_benutzer_master_ds
-- * delete_nm_benutzer_master_ds
-- können unverändert bleiben. Die variablen sind
-- dann nicht mehr selbstsprechend (benutzer_id -> gruppe_id), aber die
-- funktionlaität bleibt so erhalten. 
-- Zu beachten ist, dass die Id des Standardvertreters nun die gruppen id
-- sein muss.

-- isEditorOrStandin(integer)
-- Liefert TRUE wenn der Benutzer fuer die Fallakte 
-- a) Bearbeiter oder
-- b) Vertreter ist
-- Sonst FALSE
--
CREATE OR REPLACE FUNCTION isEditorOrStandin(mid integer)
    RETURNS boolean 
    LANGUAGE 'plpgsql' STABLE
    AS $$
DECLARE
    bid integer;
    isAllowed integer;
BEGIN
    SELECT bearbeiter_id into bid from master_tbl_view WHERE id = mid;
    IF bid = getuserid() THEN
        RETURN TRUE;
    ELSE 
        SELECT INTO isAllowed benutzer_id FROM 
        nm_benutzer_master_tbl_view 
        WHERE master_id = mid 
        AND benutzer_id IN (
                SELECT gruppe_id FROM nm_benutzer_gruppe_tbl_view 
                WHERE benutzer_id = getuserid()
                );
    END IF;
    RETURN FOUND;
END;
$$;

REVOKE ALL ON FUNCTION iseditororstandin(integer) FROM PUBLIC;
GRANT EXECUTE ON FUNCTION isEditorOrStandin(integer) TO GROUP :cm_ka_group;
GRANT EXECUTE ON FUNCTION isEditorOrStandin(integer) TO GROUP :admin_ka_group;


-- 04views
CREATE OR REPLACE VIEW ka_fall_termine_tbl_view AS 
SELECT ka_termine_tbl.id, ka_termine_tbl.master_id,
ka_termine_tbl.sdatum, ka_termine_tbl.edatum,
ka_termine_tbl.name, ka_termine_tbl.beschreibung, ka_termine_tbl.art FROM ka_termine_tbl 
INNER JOIN master_tbl ON
ka_termine_tbl.master_id = master_tbl.id 
WHERE master_tbl.bearbeiter_id = getuserid()
    OR master_tbl.id IN (
        SELECT nm.master_id from nm_benutzer_master_tbl nm
        WHERE nm.benutzer_id IN (
	    SELECT nm2.gruppe_id from nm_benutzer_gruppe_tbl nm2 
	    WHERE nm2.benutzer_id = getuserid()
	)
    )
    OR pg_has_role('admin_ka','MEMBER');

CREATE OR REPLACE VIEW ka_fall_dokumente_tbl_view AS 
SELECT ka_dokumente_tbl.id, ka_dokumente_tbl.master_id,
ka_dokumente_tbl.groesse, ka_dokumente_tbl.mime, ka_dokumente_tbl.daten,
ka_dokumente_tbl.name FROM ka_dokumente_tbl 
INNER JOIN master_tbl ON
ka_dokumente_tbl.master_id = master_tbl.id 
WHERE master_tbl.bearbeiter_id = getuserid()
    OR master_tbl.id IN (
        SELECT nm.master_id from nm_benutzer_master_tbl nm
        WHERE nm.benutzer_id IN (
	    SELECT nm2.gruppe_id from nm_benutzer_gruppe_tbl nm2 
	    WHERE nm2.benutzer_id = getuserid()
	)
    )
    OR pg_has_role('admin_ka','MEMBER');

DROP VIEW ka_benutzer_tbl_view;
CREATE OR REPLACE VIEW ka_benutzer_tbl_view AS 
SELECT ka_benutzer_tbl.id, ka_benutzer_tbl.vorname, ka_benutzer_tbl.nachname,
ka_benutzer_tbl.telefon, ka_benutzer_tbl.aktiviert,
ka_benutzer_tbl.passwort_aendern, ka_benutzer_tbl.login,
ka_benutzer_tbl.raumnummer, ka_benutzer_tbl.filiale, p.usesysid,
ka_benutzer_tbl.vertreter, getrole(ka_benutzer_tbl.id) as rolle,
ka_benutzer_tbl.gid
FROM ka_benutzer_tbl JOIN pg_user p
ON ka_benutzer_tbl.login = p.usename;

CREATE OR REPLACE RULE ka_benutzer_update_rule AS ON UPDATE TO ka_benutzer_tbl_view 
	DO INSTEAD  UPDATE ka_benutzer_tbl SET 
	vorname = new.vorname, 
	nachname = new.nachname, 
	telefon = new.telefon, 
	aktiviert = new.aktiviert, 
	raumnummer = new.raumnummer, 
	filiale = new.filiale, 
	passwort_aendern = new.passwort_aendern, 
	vertreter = new.vertreter,
	gid = new.gid
	WHERE ka_benutzer_tbl.id = new.id;

GRANT SELECT ON ka_benutzer_tbl_view TO GROUP :cm_ka_group;
GRANT SELECT ON ka_benutzer_tbl_view TO GROUP :admin_ka_group;
GRANT UPDATE ON ka_benutzer_tbl_view TO GROUP :admin_ka_group;

CREATE OR REPLACE RULE ka_benutzer_delete_rule AS ON DELETE TO ka_benutzer_tbl_view 
    DO INSTEAD NOTHING;

CREATE OR REPLACE RULE ka_benutzer_insert_rule AS ON INSERT TO ka_benutzer_tbl_view
    DO INSTEAD NOTHING;

COMMIT;
